/* This file has been prepared for Doxygen automatic documentation generation.*/
/*
 * Copyright (C) 2016 Gabriel Csefalvay
 *      Support email: tyg.res.pg@gmail.com
 *
 * Copyright (C) 2012 Yuriy Kulikov
 *      Universitaet Erlangen-Nuernberg
 *      LS Informationstechnik (Kommunikationselektronik)
 *      Support email: Yuriy.Kulikov.87@googlemail.com
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* For info on RETI instruction and PMIC operation see:
 * Any XMEGA manual > Interrupts and Programmable Multilevel Interrupt Controller > Operation
 * Any XMEGA manual > Interrupts and Programmable Multilevel Interrupt Controller > Register Description > STATUS
 *
 * If preemption is enabled, a context switch can occur in the timer interrupt. This has to be ended with RETI.
 * The interrupt is set to "low level" (lowest possible), so it can NOT interrupt any other interrupts. (This
 * would be very unsafe. :) RETI clears the LOLVLEX bit, and the PMIC returns to normal state while the
 * RETI jumps to the next task.
 * 
 * If preemption is enabled, a non-interrupt context switch could be interrupted by an in-interrupt context switch.
 * We don't want this because of higher stack usage - the context could be saved twice. The context switch can
 * occur only in the tick interrupt (and that is controlled in this file) or could in portYIELD_FROM_ISR().
 * 
 * portYIELD_FROM_ISR() is not implemented because I haven't found a solution for it so far. 
 */

/*/portYIELD_FROM_ISR() - Problem description
 *
 * The ISR prologue saves some registers. The context switch has to end with RETI to allow other interrupts to run 
 * subsequently. There will be later a context switch, that returns to this interrupted ISR. The ISR epilogue 
 * restores the registers then makes a RETI. This RETI is duplicate to the previous, so it could corrupt the 
 * internal status of PMIC.
 *
 * Possible solution: separate the tick ISR from the context switch ISR (both lo level). The context switch ISR 
 * would be fired only by software from the tick interrupt and from any other place. It could be fired from another
 * IRS as well. In this case the firing ISR would return normally, and then the context switch ISR would be executed
 * separately.
/*/


// port.c
// -------------------------------------------------------
// FreeRTOS 8.2.3 port for Atmel XMEGA microcontrollers


//  version  date        descr.
// ------------------------------------------------------------
//    v00    28.05.2016  initial
//    v01    25.07.2016  more description
//    v02    10.08.2016  clear the RAMPZ after it's saved


#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>
#include "FreeRTOSConfig.h"
#include "../../FreeRTOS.h"
#include "task.h"


#ifndef configPORT_DISABLE_ALL_INTERRUPTS_DURING_CONTEXT_SWITCH
  #define configPORT_DISABLE_ALL_INTERRUPTS_DURING_CONTEXT_SWITCH  0
#endif

#ifndef configPORT_SAVE_RAMPZ_ON_CONTEXT_SWITCH          
  #define configPORT_SAVE_RAMPZ_ON_CONTEXT_SWITCH  0
#endif

#ifndef configPORT_TICK_TIMER
  #define configPORT_TICK_TIMER  E0
  #ifdef configPORT_TICK_TIMER_INTCTRLA_ADDRESS
    #undef configPORT_TICK_TIMER_INTCTRLA_ADDRESS
  #endif
  #define configPORT_TICK_TIMER_INTCTRLA_ADDRESS  0x0A06
#endif

#ifndef configPORT_TICK_TIMER_INTCTRLA_ADDRESS
  #error You have to also define configPORT_TICK_TIMER_INTCTRLA_ADDRESS!
#endif


#define ever (;;)
#define CATx(x,y)  x ## y
#define CAT(x,y)   CATx(x,y)
#define STR(s) STRx(s)
#define STRx(s) #s

#define TC_TICK    CAT(TC, configPORT_TICK_TIMER)
#define TC_ISR_vect_2  CAT(configPORT_TICK_TIMER, _OVF_vect)
#define TC_ISR_vect    CAT(TC, TC_ISR_vect_2)

#define portSREG_FLAGS_INT_ENABLED  ((StackType_t)CPU_I_bm)
#define TC_TICK_INT_gc              TC_OVFINTLVL_LO_gc
/*-----------------------------------------------------------*/

/* We require the address of the pxCurrentTCB variable, but don't want to know
 any details of its type. */
typedef void TCB_t;
extern volatile TCB_t * volatile pxCurrentTCB;

/*-----------------------------------------------------------*/

/* 
 * Macros to save all the general purpose registers, then save the stack pointer
 * into the TCB.  
 * 
 * if configPORT_DISABLE_ALL_INTERRUPTS_DURING_CONTEXT_SWITCH == 1, interrupts 
 * are globally disabled when switching context.
 * 
 * if configPORT_DISABLE_ALL_INTERRUPTS_DURING_CONTEXT_SWITCH == 0:
 * - if configUSE_PREEMPTION == 1, the tick interrupt is disabled when switching
 *   context.
 * - if configUSE_PREEMPTION == 0, no interrupts are disabled.
 * - if configPORT_SAVE_RAMPZ_ON_CONTEXT_SWITCH == 1, the RAMPZ register is also 
 *   saved to the stack and cleared.    
 */



#define portSAVE_CONTEXT()          { portSAVE_CONTEXT_R31_SREG_INT();     portSAVE_CONTEXT_R0_R30(); portSAVE_CONTEXT_RAMPZ(); portSAVE_CONTEXT_STACK_POINTER(); }
#define portSAVE_CONTEXT_IN_TICK()  { portSAVE_CONTEXT_R31_SREG_IN_TICK(); portSAVE_CONTEXT_R0_R30(); portSAVE_CONTEXT_RAMPZ(); portSAVE_CONTEXT_STACK_POINTER(); }

#if configPORT_DISABLE_ALL_INTERRUPTS_DURING_CONTEXT_SWITCH

  #define portSAVE_CONTEXT_R31_SREG_INT()      portSAVE_CONTEXT_R31_SREG_CLI()
  #define portSAVE_CONTEXT_R31_SREG_IN_TICK()  portSAVE_CONTEXT_R31_SREG_CLI()
  
#else

  #define portSAVE_CONTEXT_R31_SREG_IN_TICK()  portSAVE_CONTEXT_R31_SREG()

  #if configUSE_PREEMPTION==1
    #define portSAVE_CONTEXT_R31_SREG_INT() portSAVE_CONTEXT_R31_SREG_DISTICKINT()
  #else
    #define portSAVE_CONTEXT_R31_SREG_INT() portSAVE_CONTEXT_R31_SREG()
    /* when no preemption, tick interrupt doesn't have to be turned off. No context switch
     * will occur in interrupts because portYIELD_FROM_ISR() is not implemented */
  #endif
  
#endif


#define portSAVE_CONTEXT_R31_SREG_DISTICKINT()        \
  asm volatile (                                      \
    "push    r31                     \n\t"            \
    "ldi     r31, %[INTLVL]          \n\t"            \
    "sts   " STR(configPORT_TICK_TIMER_INTCTRLA_ADDRESS) ", r31  \n\t" \
    "in      r31, __SREG__           \n\t"            \
    "push    r31                     \n\t"            \
    :                                                 \
    : [INTLVL] "M" ((uint8_t)TC_OVFINTLVL_OFF_gc)     \
    : "r31"                                           \
  )
_Static_assert((uint16_t)&TC_TICK.INTCTRLA == (uint16_t)configPORT_TICK_TIMER_INTCTRLA_ADDRESS,
               "configPORT_TICK_TIMER_INTCTRLA_ADDRESS does not correspont to INTCTRLA register of configPORT_TICK_TIMER");

#define portSAVE_CONTEXT_R31_SREG_CLI()       \
  asm volatile (                              \
    "push    r31                     \n\t"    \
    "in      r31, __SREG__           \n\t"    \
    "cli                             \n\t"    \
    "push    r31                     \n\t"    \
    ::: "r31"                                 \
  )

#define portSAVE_CONTEXT_R31_SREG()           \
  asm volatile (                              \
    "push    r31                     \n\t"    \
    "in      r31, __SREG__           \n\t"    \
    "push    r31                     \n\t"    \
    ::: "r31"                                 \
  )

#define portSAVE_CONTEXT_R0_R30()  \
  asm volatile (            \
    "push    r0                     \n\t"    \
    "push    r1                     \n\t"    \
    "clr     r1                     \n\t"    \
    "push    r2                     \n\t"    \
    "push    r3                     \n\t"    \
    "push    r4                     \n\t"    \
    "push    r5                     \n\t"    \
    "push    r6                     \n\t"    \
    "push    r7                     \n\t"    \
    "push    r8                     \n\t"    \
    "push    r9                     \n\t"    \
    "push    r10                    \n\t"    \
    "push    r11                    \n\t"    \
    "push    r12                    \n\t"    \
    "push    r13                    \n\t"    \
    "push    r14                    \n\t"    \
    "push    r15                    \n\t"    \
    "push    r16                    \n\t"    \
    "push    r17                    \n\t"    \
    "push    r18                    \n\t"    \
    "push    r19                    \n\t"    \
    "push    r20                    \n\t"    \
    "push    r21                    \n\t"    \
    "push    r22                    \n\t"    \
    "push    r23                    \n\t"    \
    "push    r24                    \n\t"    \
    "push    r25                    \n\t"    \
    "push    r26                    \n\t"    \
    "push    r27                    \n\t"    \
    "push    r28                    \n\t"    \
    "push    r29                    \n\t"    \
    "push    r30                    \n\t"    \
  )

#if configPORT_SAVE_RAMPZ_ON_CONTEXT_SWITCH
  #define portSAVE_CONTEXT_RAMPZ()            \
    asm volatile (                            \
      "in      r0, __RAMPZ__           \n\t"  \
      "push    r0                      \n\t"  \
      "out     __RAMPZ__, __zero_reg__ \n\t"  \
    )
 #else
  #define portSAVE_CONTEXT_RAMPZ()
#endif

#define portSAVE_CONTEXT_STACK_POINTER()     \
  asm volatile (                             \
    "lds     r26, pxCurrentTCB      \n\t"    \
    "lds     r27, pxCurrentTCB + 1  \n\t"    \
    "in      r0, __SP_L__           \n\t"    \
    "st      x+, r0                 \n\t"    \
    "in      r0, __SP_H__           \n\t"    \
    "st      x+, r0                 \n\t"    \
    ::: "r26", "r27"                         \
  )


/* 
 * Opposite to portSAVE_CONTEXT().  Interrupts will have been disabled during
 * the context save so we can write to the stack pointer. 
 */


#define portRESTORE_CONTEXT()          { portRESTORE_CONTEXT_STACK_POINTER(); portRESTORE_CONTEXT_RAMPZ(); portRESTORE_CONTEXT_R30_R0(); portRESTORE_CONTEXT_SREG_R31_INT(); }
#define portRESTORE_CONTEXT_IN_TICK()  { portRESTORE_CONTEXT_STACK_POINTER(); portRESTORE_CONTEXT_RAMPZ(); portRESTORE_CONTEXT_R30_R0(); portRESTORE_CONTEXT_SREG_R31(); }

#if configPORT_DISABLE_ALL_INTERRUPTS_DURING_CONTEXT_SWITCH==0 && configUSE_PREEMPTION==1
  #define portRESTORE_CONTEXT_SREG_R31_INT()  portRESTORE_CONTEXT_SREG_R31_DISTICKINT()
#else
  #define portRESTORE_CONTEXT_SREG_R31_INT()  portRESTORE_CONTEXT_SREG_R31()
  /* if configPORT_DISABLE_INTERRUPTS_DURING_CONTEXT_SWITCH==1 then I bit is restored in the SREG so no special instructions are required. */
  /* if configPORT_DISABLE_INTERRUPTS_DURING_CONTEXT_SWITCH==0 && configUSE_PREEMPTION==0 then no interrupts are disabled */
#endif

#define portRESTORE_CONTEXT_STACK_POINTER() \
  asm volatile (              \
    "lds    r26, pxCurrentTCB        \n\t"    \
    "lds    r27, pxCurrentTCB + 1    \n\t"    \
    "ld     r28, x+                  \n\t"    \
    "ld     r29, x+                  \n\t"    \
    "out    __SP_L__, r28            \n\t"  /* see note below */\
    "out    __SP_H__, r29            \n\t"    \
  )
/* "out __SP_L__, r28" disables interrupts automatically until the next out instr.. See Any XMEGA Manual > Atmel AVR CPU > Register Descriptions > SPL */

#if configPORT_SAVE_RAMPZ_ON_CONTEXT_SWITCH
  #define portRESTORE_CONTEXT_RAMPZ() \
    asm volatile (              \
      "pop    r0                       \n\t"    \
      "out    __RAMPZ__, r0            \n\t"    \
    )
#else
  #define portRESTORE_CONTEXT_RAMPZ()
#endif

#define portRESTORE_CONTEXT_R30_R0()          \
  asm volatile (                              \
    "pop    r30                      \n\t"    \
    "pop    r29                      \n\t"    \
    "pop    r28                      \n\t"    \
    "pop    r27                      \n\t"    \
    "pop    r26                      \n\t"    \
    "pop    r25                      \n\t"    \
    "pop    r24                      \n\t"    \
    "pop    r23                      \n\t"    \
    "pop    r22                      \n\t"    \
    "pop    r21                      \n\t"    \
    "pop    r20                      \n\t"    \
    "pop    r19                      \n\t"    \
    "pop    r18                      \n\t"    \
    "pop    r17                      \n\t"    \
    "pop    r16                      \n\t"    \
    "pop    r15                      \n\t"    \
    "pop    r14                      \n\t"    \
    "pop    r13                      \n\t"    \
    "pop    r12                      \n\t"    \
    "pop    r11                      \n\t"    \
    "pop    r10                      \n\t"    \
    "pop    r9                       \n\t"    \
    "pop    r8                       \n\t"    \
    "pop    r7                       \n\t"    \
    "pop    r6                       \n\t"    \
    "pop    r5                       \n\t"    \
    "pop    r4                       \n\t"    \
    "pop    r3                       \n\t"    \
    "pop    r2                       \n\t"    \
    "pop    r1                       \n\t"    \
    "pop    r0                       \n\t"    \
  )

#define portRESTORE_CONTEXT_SREG_R31() \
  asm volatile (              \
    "pop    r31                       \n\t"    \
    "out    __SREG__, r31             \n\t"    \
    "pop    r31                       \n\t"    \
  )

#define portRESTORE_CONTEXT_SREG_R31_DISTICKINT()    \
  asm volatile (                                     \
    "pop    r31                      \n\t"           \
    "out    __SREG__, r31            \n\t"           \
    "ldi     r31, %[INTLVL]          \n\t"           \
    "sts   " STR(configPORT_TICK_TIMER_INTCTRLA_ADDRESS) ", r31  \n\t" \
    "pop    r31                      \n\t"           \
    :                                                \
    : [INTLVL]  "M" ((uint8_t)TC_TICK_INT_gc)        \
  )

/*-----------------------------------------------------------*/

/*
 * Perform hardware setup to enable ticks from a timer
 */
static void prvSetupTickInterrupt(void);


/*-----------------------------------------------------------*/

/* 
 * See header file for description. 
 */

StackType_t *pxPortInitialiseStack( StackType_t *pxTopOfStack, TaskFunction_t pxCode, void *pvParameters )
{
    /*The addresses are 16 or 24 bit depending on the xmega memory, so use 32 bit variable but put only a
     * part to stack. This means we use can only 16bit addresses. */
    
    uint16_t usAddress;
    
    /* Place a few bytes of known values on the bottom of the stack.
     This is just useful for debugging. */
    
    *pxTopOfStack = 0x11;
    pxTopOfStack--;
    *pxTopOfStack = 0x22;
    pxTopOfStack--;
    *pxTopOfStack = 0x33;
    pxTopOfStack--;
    
    /* Simulate how the stack would look after a call to vPortYield() generated by
     the compiler. */
    /*lint -e950 -e611 -e923 Lint doesn't like this much - but nothing I can do about it. */
    /* The start of the task code will be popped off the stack last, so place
     it on first. */
    

    /* For XMEGAs, avr-gcc generates valid 16-bit function pointers for every function (trampoline)
     * so the 3rd address byte is always 0. */

    usAddress = (uint16_t)pxCode;

    *pxTopOfStack = (StackType_t ) (usAddress & 0x00ff);
    pxTopOfStack--;

    usAddress >>= 8;
    *pxTopOfStack = (StackType_t ) (usAddress & 0x00ff);
    pxTopOfStack--;

    #if defined(__AVR_3_BYTE_PC__) && __AVR_3_BYTE_PC__
      *pxTopOfStack = (StackType_t ) 0;
      pxTopOfStack--;
    #endif

    /* Next simulate the stack as if after a call to portSAVE_CONTEXT().
     portSAVE_CONTEXT places the flags on the stack immediately after r0
     to ensure the interrupts get disabled as soon as possible, and so ensuring
     the stack use is minimal should a context switch interrupt occur. */
    
    *pxTopOfStack = (StackType_t ) 0x31; /* R31 */
    pxTopOfStack--;
    
    *pxTopOfStack = portSREG_FLAGS_INT_ENABLED;
    pxTopOfStack--;
    
    /* Now the remaining registers.   The compiler expects R1 to be 0. */
    
    *pxTopOfStack = (StackType_t ) 0x32; /* R0 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x00; /* R1 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x02; /* R2 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x03; /* R3 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x04; /* R4 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x05; /* R5 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x06; /* R6 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x07; /* R7 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x08; /* R8 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x09; /* R9 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x10; /* R10 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x11; /* R11 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x12; /* R12 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x13; /* R13 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x14; /* R14 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x15; /* R15 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x16; /* R16 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x17; /* R17 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x18; /* R18 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x19; /* R19 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x20; /* R20 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x21; /* R21 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x22; /* R22 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x23; /* R23 */
    pxTopOfStack--;
    
    /* Place the custom parameter pointer (pvParameters) on the stack in the expected location. */
    
    usAddress = (uint16_t) pvParameters;
    *pxTopOfStack = (StackType_t ) (usAddress & (uint16_t) 0x00ff); /* R24 */
    pxTopOfStack--;

    usAddress >>= 8;
    *pxTopOfStack = (StackType_t ) (usAddress & (uint16_t) 0x00ff); /* R25 */
    pxTopOfStack--;

    *pxTopOfStack = (StackType_t ) 0x26; /* R26 X */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x27; /* R27 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x28; /* R28 Y */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x29; /* R29 */
    pxTopOfStack--;
    *pxTopOfStack = (StackType_t ) 0x30; /* R30 Z */
    pxTopOfStack--;
    
    /* RAMPZ - default value is 0x00 */

    #if configPORT_SAVE_RAMPZ_ON_CONTEXT_SWITCH
      *pxTopOfStack = (StackType_t ) 0x00; /* RAMPZ */
      pxTopOfStack--;
    #endif

    return pxTopOfStack;
}

/*-----------------------------------------------------------*/

BaseType_t xPortStartScheduler(void)
{
    /* Setup the hardware to generate the tick. */
    prvSetupTickInterrupt();

    /* Restore the context of the first task that is going to run. */
    portRESTORE_CONTEXT();

    /* Simulate a function call end as generated by the compiler.  We will now
     jump to the start of the task the context of which we have just restored. */
    asm volatile ( "ret" );

    /* Should not get here. */
    return pdTRUE;
}

/*-----------------------------------------------------------*/

void vPortEndScheduler(void) 
{
  for ever;
}

/*-----------------------------------------------------------*/

/*
 * Manual context switch.
 */
void vPortYield(void) __attribute__ ((naked));
void vPortYield(void)
{
    portSAVE_CONTEXT();
    vTaskSwitchContext();
    portRESTORE_CONTEXT();
    asm volatile ( "ret" );
}

/*-----------------------------------------------------------*/


/*
 * Setup a timer to generate a tick interrupt.
 */

static void prvSetupTickInterrupt(void) 
{
  TC_TICK.PER = configCPU_CLOCK_HZ / configTICK_RATE_HZ / 64 - 1;
  TC_TICK.CTRLA = TC_CLKSEL_DIV64_gc;
  TC_TICK.INTCTRLA = TC_TICK_INT_gc;
}

/*-----------------------------------------------------------*/

#if configUSE_PREEMPTION == 1

/*
 * Tick ISR for preemptive scheduler.  We use the naked attribute as
 * the context is saved at the start and restored at the end manually.
 * The tick count is incremented after the context is saved. Context 
 * is switched only when needed.
 *
 * This must be identical to vPortYield() from the call to 
 * vTaskSwitchContext() onwards. The only difference from vPortYield() 
 * is the tick count is incremented as the call comes from the 
 * tick ISR.
 */

ISR (TC_ISR_vect, ISR_NAKED)
{
  portSAVE_CONTEXT_IN_TICK();
	if( xTaskIncrementTick() != pdFALSE )
		vTaskSwitchContext();
	portRESTORE_CONTEXT_IN_TICK();
  asm volatile ( "reti" ); /* see notes at top on reti */
}

#else

/*
 * Tick ISR for the cooperative scheduler.  All this does is increment the
 * tick count.  We don't need to switch context, this can only be done by
 * manual calls to taskYIELD();
 */

ISR(TC_ISR_vect)
{
  xTaskIncrementTick();
}

#endif

