// (C) 2016, Gabriel Csefalvay

#include "procutil.h"

#include <avr/io.h>



// -------------------
// --- oscillators ---
// -------------------



// Enable the external 32.768kHz crystal
// -------------------------------------

void procutil_clk_xtal_32kiHz_en(void)
{
  OSC.XOSCCTRL = OSC_XOSCSEL_32KHz_gc; // crystal type: 32.768kHz
  OSC.CTRL |= OSC_XOSCEN_bm; // enable XOSC
}


// Enable the internal 32.768kHz osc.
// ----------------------------------

void procutil_clk_RC32kiHz_en(void)
{
  OSC.CTRL |= OSC_RC32KEN_bm; // 32.768kHz int. ref
}


// Enable the RC32MHz osc.
// --------------------------------------

void procutil_clk_RC32MHz_en(void)
{
  OSC.CTRL |= OSC_RC32MEN_bm; // turn on 32MHz internal osc.
}


// Enable the DFLL module on RC32MHz osc.
// Nominal reference clock is 1024Hz.
// --------------------------------------
// The reference clock has to be enabled
// prior to call.
// --------------------------------------
// input
//   compensation_promile: frequency compensation
//     + for slower reference or faster output clock
//     - for faster reference or slower output clock
//  source: select the reference clock source

void procutil_clk_RC32MHz_DFLL_en(clk_dfll_conf_t conf)
{
  // select ref. source
  // wait for ready
  
  switch (conf.source)
  {
    case clk_dfll_RC32kiHz:
      while (! (OSC.STATUS & OSC_RC32KRDY_bm) );
      
      #if defined (OSC_RC32MCREF_bm)
        OSC.DFLLCTRL &= ~OSC_RC32MCREF_bm;
      #elif defined (OSC_RC32MCREF_gm)
        OSC.DFLLCTRL &= ~OSC_RC32MCREF_gm;
        OSC.DFLLCTRL |= OSC_RC32MCREF_RC32K_gc;
      #else 
        #error clk.h: clk_RC32MHz_DFLL_en(): OSC.DFLLCTRL content definition problem.
      #endif
      
      break;
    
    case clk_dfll_xtal_32kiHz:
      while (! (OSC.STATUS & OSC_XOSCRDY_bm) );

      #if defined (OSC_RC32MCREF_bm)
        OSC.DFLLCTRL |= OSC_RC32MCREF_bm;
      #elif defined (OSC_RC32MCREF_gm)
        OSC.DFLLCTRL &= ~OSC_RC32MCREF_gm;
        OSC.DFLLCTRL |= OSC_RC32MCREF_XOSC32K_gc;
      #endif
      
      break;
  }
  
  while (! (OSC.STATUS & OSC_RC32MRDY_bm) );
    
  // enable DFLL, 32MHz/1024Hz = 31250 (0x7A12)
  
  uint16_t div = 31250;
  int16_t  comp = (int32_t)div * conf.compensation_promile / 1000;
  
  div += comp;
  
  DFLLRC32M.COMP2 = (div>>8) & 0xFF;
  DFLLRC32M.COMP1 = (div>>0) & 0xFF;
  DFLLRC32M.CTRL = DFLL_ENABLE_bm;
}


// Enable the RC32MHz osc.
// -----------------------

void procutil_clk_RC32MHz_dis(void)
{
  OSC.CTRL &= ~OSC_RC32MEN_bm; // turn off 32MHz internal osc.
}



// --------------------------------
// --- clock source and divider ---
// --------------------------------



// 32MHz clock from RC32MHz osc.
// The osc. has to be enabled prior
// to call.
// --------------------------------

void procutil_clk_32MHz_from_RC32MHz(void)
{
  // wait for ready
  while (! (OSC.STATUS & OSC_RC32MRDY_bm) ); 

  // set divider: 32MHz/1 = 32MHz
  CCP = CCP_IOREG_gc;
  CLK.PSCTRL = CLK_PSADIV_1_gc | CLK_PSBCDIV_1_1_gc; 

  // switch to RC32MHz
  CCP = CCP_IOREG_gc;
  CLK.CTRL = CLK_SCLKSEL_RC32M_gc;              
}


// 8MHz clock from RC32MHz osc.
// The osc. has to be enabled prior
// to call.
// --------------------------------

void procutil_clk_8MHz_from_RC32MHz(void)
{
  // wait for ready
  while (! (OSC.STATUS & OSC_RC32MRDY_bm) );

  // set divider: 32MHz/4 = 32MHz
  CCP = CCP_IOREG_gc;
  CLK.PSCTRL = CLK_PSADIV_4_gc | CLK_PSBCDIV_1_1_gc;

  // switch to RC32MHz
  CCP = CCP_IOREG_gc;
  CLK.CTRL = CLK_SCLKSEL_RC32M_gc;
}


// 2MHz clock from RC2MHz osc.
// The osc. has to be enabled prior
// to call.
// --------------------------------

void procutil_clk_2MHz_from_RC2MHz(void)
{
  // wait for ready
  while (! (OSC.STATUS & OSC_RC32MRDY_bm) );

  // switch to RC2MHz
  CCP = CCP_IOREG_gc;
  CLK.CTRL = CLK_SCLKSEL_RC2M_gc;

  // set divider: 2MHz/1 = 2MHz
  CCP = CCP_IOREG_gc;
  CLK.PSCTRL = CLK_PSADIV_1_gc | CLK_PSBCDIV_1_1_gc;
}


// Enable all interrupt types
// in the PMIC
// ---------------------------------

void procutil_int_pmic_enable_all(void)
{
  PMIC.CTRL |= PMIC_HILVLEN_bm | PMIC_MEDLVLEN_bm | PMIC_LOLVLEN_bm;
}